@@ -4,6 +4,7 @@ from __future__ import division, absolute_import, print_function, unicode_litera
 
 import logging
 import re
+import warnings
 from calendar import timegm
 from datetime import datetime, timedelta
 from email.utils import parsedate_tz
@@ -11,24 +12,20 @@ from email.utils import parsedate_tz
 try: import ujson as json
 except ImportError: import json
 
-try: unicode # Python 2
-except NameError: unicode = str # Python 3
-else: # Python 2
-	def bytes(s, encoding=None, errors=None):
-		return s.encode(encoding, errors)
+from .connection_pool import pool
 
 try: # Python 3
-	from urllib.request import urlopen
-	from urllib.parse import urlencode
 	from urllib.error import HTTPError
+	from urllib.parse import urlencode
+	unicode = basestring = str
 
 except ImportError: # Python 2
-	from urllib import urlopen, urlencode
-	
-	class HTTPError(Exception): pass
+	from urllib import urlencode
+	from urllib2 import HTTPError
+	def bytes(s, encoding=None, errors=None): return s.encode(encoding, errors)
 
 
-VERSION = (1, 5, 2)
+VERSION = (1, 6, 0)
 
 __title__ = 'Chump'
 __version__ = '.'.join((unicode(i) for i in VERSION))
@@ -40,76 +37,51 @@ __copyright__ = 'Copyright 2013 Karan Lyons'
 
 
 logger = logging.getLogger(__name__)
-
-try:
-	logger.addHandler(logging.NullHandler())
-
-except AttributeError:
-	class NullHandler(logging.Handler):
-		def emit(self, record):
-			pass
-	
-	logger.addHandler(NullHandler())
+logger.addHandler(logging.NullHandler())
 
 
-try:
+try: # pytz installed
 	from pytz import utc
+	def utc_now(): return utc.localize(datetime.utcnow())
+	def epoch_to_datetime(e): return utc.localize(datetime.utcfromtimestamp(int(e)))
 
 except ImportError:
-	logger.warning('pytz is not installed; datetime\'s may be inaccurate.')
+	try: # dateutil installed
+		from dateutil.tz import tzutc
+		utc = tzutc()
 	
-	from datetime import tzinfo
+	except ImportError:
+		try: # Python >= 3.2
+			from datetime import timezone
+			utc = timezone.utc
+		
+		except ImportError: # Python < 3.2
+			from datetime import tzinfo
+			class UTC(tzinfo):
+				ZERO = timedelta(0)
+				def utcoffset(self, dt): return self.ZERO
+				def tzname(self, dt): return 'UTC'
+				def dst(self, dt): return self.ZERO
+				def __unicode__(self): return 'UTC'
+				__str__ = __unicode__
+				def __repr__(self): return 'chump.UTC'
+			utc = UTC()
 	
-	class UTC(tzinfo):
-		ZERO = timedelta(0)
-		
-		def utcoffset(self, dt):
-			return self.ZERO
-		
-		def tzname(self, dt):
-			return 'UTC'
-		
-		def dst(self, dt):
-			return self.ZERO
-		
-		def localize(self, dt):
-			return dt.replace(tzinfo=self)
-		
-		def __unicode__(self):
-			return 'UTC'
-		
-		__str__ = __unicode__
-		
-		def __repr__(self):
-			return '<UTC>'
-	
-	utc = UTC()
-
-
-def utc_now():
-	return utc.localize(datetime.utcnow())
+	def utc_now(): return datetime.utcnow().replace(tzinfo=utc)
+	def epoch_to_datetime(e): return datetime.utcfromtimestamp(int(e)).replace(tzinfo=utc)
 
 
 def datetime_to_epoch(dt):
-	if dt.tzinfo is None: # We got a naïve datetime. Assume it's UTC.
-		dt = utc.localize(dt)
-	
-	return timegm(dt.astimezone(utc).timetuple())
-
-
-def epoch_to_datetime(e):
-	return utc.localize(datetime.utcfromtimestamp(int(e)))
+	if dt.tzinfo is not None: dt = dt.astimezone(utc)
+	else: warnings.warn('Naïve datetime received: assuming UTC', RuntimeWarning)
+	return timegm(dt.timetuple())
 
 
 def http_date_to_datetime(d):
 	d_tuple = parsedate_tz(d)
 	
-	dt = utc.localize(datetime(*d_tuple[:6]))
-	
-	if d_tuple[9]: # We've got a timezone offset.
-		dt += timedelta(seconds=d_tuple[9])
-	
-	return dt
+	if d_tuple is not None:
+		return datetime(*d_tuple[:6]).replace(tzinfo=utc) - timedelta(seconds=d_tuple[9])
 
 
 LOWEST = -2 #: Message priority: No sound, no vibration, no banner.
@@ -120,21 +92,26 @@ EMERGENCY = 2 #: Message priority: Sound, vibration, and banner regardless of us
 
 
 TOKEN_RE = re.compile(r'^[a-zA-Z0-9]{30}$') # Matches correct application/user tokens.
+DEVICE_RE = re.compile(r'^[A-Za-z0-9_-]{,25}$') # Matches correct device names.
 
 
 ENDPOINT = 'https://api.pushover.net/1/'
 REQUESTS = {
-	'message': {
-		'method': 'post',
-		'path': 'messages.json',
+	'sound': {
+		'method': 'get',
+		'path': 'sounds.json',
+	},
+	'limits': {
+		'method': 'get',
+		'path': 'apps/limits.json',
 	},
 	'validate': {
 		'method': 'post',
 		'path': 'users/validate.json',
 	},
-	'sound': {
-		'method': 'get',
-		'path': 'sounds.json',
+	'message': {
+		'method': 'post',
+		'path': 'messages.json',
 	},
 	'receipt': {
 		'method': 'get',
@@ -203,23 +180,65 @@ class Application(object):
 	"""
 	
 	def __init__(self, token):
-		self.is_authenticated = False #: A :py:obj:`bool` indicating whether the application has been authenticated.
-		self.sounds = None #: If authenticated, a :py:class:`dict` of available notification sounds, otherwise :py:obj:`None`.
-		self.token = unicode(token) #: A :py:obj:`string` of the application's API token.
+		self.token = token #: A :py:obj:`string` of the application's API token.
+		self._is_authenticated = None
+		self._sounds = None
 		
 		self.limit = None #: If a message has been sent, an :py:obj:`int` of the application's monthly message limit, otherwise :py:obj:`None`.
 		self.remaining = None #: If a message has been sent, an :py:obj:`int` of the application's remaining message allotment, otherwise :py:obj:`None`.
 		self.reset = None #: If a message has been sent, :py:class:`~datetime.datetime` of when the application's monthly message limit will reset, otherwise :py:obj:`None`.
 	
+	@property
+	def is_authenticated(self):
+		"""
+		A lazily loaded :py:obj:`bool` indicating whether the application
+		is authenticated.
+		
+		"""
+		
+		if self._is_authenticated is None:
+			self._authenticate()
+		
+		return self._is_authenticated
+	
+	@is_authenticated.setter
+	def is_authenticated(self, value):
+		self._is_authenticated = value
+	
+	@property
+	def sounds(self):
+		"""
+		A lazily loaded :py:class:`dict` of available notification sounds if
+		authenticated, otherwise :py:obj:`None`.
+		
+		"""
+		
+		if self._sounds is None and self._is_authenticated is not False:
+			self._authenticate()
+		
+		return self._sounds
+	
+	@sounds.setter
+	def sounds(self, value):
+		self._sounds = value
+	
 	def __setattr__(self, name, value):
+		if name == 'token':
+			try:
+				if not TOKEN_RE.match(value):
+					raise ValueError('Bad application token: expected string matching r{pattern!r}, got {value!r}'.format(pattern=TOKEN_RE.pattern, value=value))
+				
+				else:
+					old_token = getattr(self, 'token', None)
+			
+			except TypeError:
+				raise TypeError('Bad token: expected string, got {value_type}'.format(value_type=type(value)))
+			
 		super(Application, self).__setattr__(name, value)
 		
-		if name == 'token':
-			if not TOKEN_RE.match(value):
-				raise ValueError('Bad application token: expected string matching [a-zA-Z0-9]{{30}}, got {value!r}'.format(value=value))
-			
-			else:
-				self._authenticate()
+		if name == 'token' and self.token != old_token:
+			super(Application, self).__setattr__('_is_authenticated', None)
+			super(Application, self).__setattr__('_sounds', None)
 	
 	def __unicode__(self):
 		return "Pushover Application: {token}".format(token=self.token)
@@ -242,24 +261,23 @@ class Application(object):
 	
 	def _authenticate(self):
 		"""
-		Authenticates the supplied application token.
+		Authenticates the supplied application token and populates available
+		notification sounds.
 		
 		"""
 		
-		self.is_authenticated = True
-		
-		# There's actually no nice way to do this in the API, so we instead try
-		# to validate a user with no key, and check to see if the returned
-		# error also includes an error about our token.
+		# We'll make a request for sounds (which we need to make regardless),
+		# and if that error fails with a token error we know we're
+		# unauthenticated.
 		try:
-			self._request('validate')
+			self._sounds = self._request('sound')[0]['sounds']
 		
 		except APIError as error:
 			if 'token' in error.bad_inputs:
-				self.is_authenticated = False
+				self._is_authenticated = False
 		
-		if self.is_authenticated and self.sounds is None:
-			self.sounds = self._request('sound')[0]['sounds']
+		else:
+			self._is_authenticated = True
 	
 	def get_user(self, token):
 		"""
@@ -313,14 +331,14 @@ class Application(object):
 				if data:
 					url += '?' + urlencode(data)
 				
-				response = urlopen(url)
+				response = pool.open(url)
 			
 			elif method == 'post':
-				response = urlopen(url, bytes(urlencode(data), 'utf-8', 'strict') if data else None)
+				response = pool.open(url, bytes(urlencode(data), 'utf-8', 'strict') if data else None)
 		
-		except HTTPError as e: # Python 3 on "error" status codes.
-			response = e
-			response.__dict__['headers'] = response.hdrs
+		except HTTPError as error_response:
+			response = error_response
+			response.__dict__['headers'] = error_response.hdrs
 		
 		response.content = response.read().decode()
 		
@@ -367,19 +385,60 @@ class User(object):
 	
 	def __init__(self, app, token):
 		self.app = app #: The Pushover application to send messages with.
-		self.is_authenticated = None #: If :attr:`.app` has been authenticated, a :py:obj:`bool` indicating whether the user has been authenticated, otherwise :py:obj:`None`.
-		self.devices = None #: If authenticated, a :py:class:`set` of the user's devices, otherwise :py:obj:`None`.
-		self.token = unicode(token) #: A :py:obj:`string` of the user's API token.
+		self.token = token #: A :py:obj:`string` of the user's API token.
+		self._is_authenticated = None
+		self._devices = None
+	
+	@property
+	def is_authenticated(self):
+		"""
+		A lazily loaded :py:obj:`bool` indicating whether the user is authenticated.
+		
+		"""
+		
+		if self._is_authenticated is None and self.app._is_authenticated is not False:
+			self._authenticate()
+		
+		return self.app._is_authenticated is True and self._is_authenticated
+	
+	@is_authenticated.setter
+	def is_authenticated(self, value):
+		self._is_authenticated = value
+	
+	@property
+	def devices(self):
+		"""
+		A lazily loaded a :py:class:`set` of the user's devices if authenticated,
+		otherwise :py:obj:`None`.
+		
+		"""
+		
+		if self._devices is None and self._is_authenticated is not False and self.app._is_authenticated is not False:
+			self._authenticate()
+		
+		return self._devices
+	
+	@devices.setter
+	def devices(self, value):
+		self._devices = value
 	
 	def __setattr__(self, name, value):
+		try:
+			if name == 'token':
+				if not TOKEN_RE.match(value):
+					raise ValueError('Bad user token: expected string matching r{pattern!r}, got {value!r}'.format(pattern=TOKEN_RE.pattern, value=value))
+				
+				else:
+					old_token = getattr(self, 'token', None)
+		
+		except TypeError:
+			raise TypeError('Bad token: expected string, got {value_type}'.format(value_type=type(value)))
+		
 		super(User, self).__setattr__(name, value)
 		
-		if name == 'token':
-			if not TOKEN_RE.match(value):
-				raise ValueError('Bad user token: expected string matching [a-zA-Z0-9]{{30}}, got {value!r}'.format(value=value))
-			
-			else:
-				self._authenticate()
+		if name == 'token' and self.token != old_token:
+			super(User, self).__setattr__('_is_authenticated', None)
+			super(User, self).__setattr__('_devices', None)
 	
 	def __unicode__(self):
 		return "Pushover User: {token}".format(token=self.token)
@@ -411,21 +470,26 @@ class User(object):
 		
 		except APIError as error:
 			if 'token' in error.bad_inputs: # We can't authenticate users with a bad API token.
-				self.app.is_authenticated = False
-				self.is_authenticated = None
-				self.devices = None
-			
-			elif 'user' not in error.bad_inputs or error.bad_inputs['user'].startswith('valid'):
-				self.is_authenticated = True
-				self.devices = set()
+				self.app._is_authenticated = False
+				self.app._sounds = None
+				self._is_authenticated = None
+				self._devices = None
 			
 			else:
-				self.is_authenticated = False
-				self.devices = None
+				self.app._is_authenticated = True
+				
+				if 'user' not in error.bad_inputs or error.bad_inputs['user'].startswith('valid'):
+					self._is_authenticated = True
+					self._devices = set()
+				
+				else:
+					self._is_authenticated = False
+					self._devices = None
 		
 		else:
-			self.is_authenticated = True
-			self.devices = set(response['devices'])
+			self.app._is_authenticated = True
+			self._is_authenticated = True
+			self._devices = set(response['devices'])
 	
 	def create_message(self, message, html=False, title=None, timestamp=None,
 		               url=None, url_title=None, device=None, priority=NORMAL,
@@ -561,78 +625,83 @@ class Message(object):
 	__le__ = __gt__ = __ge__ = __lt__
 	
 	def __setattr__(self, name, value):
-		if name == 'message' and len(value) == 0:
-			raise ValueError('Bad message: must be > 0 characters, was 0')
-		
 		if name == 'html':
 			try:
-				value_type = type(value)
 				value = int(value)
 			
 			except ValueError:
-				raise ValueError('Bad html: expected bool, got {value_type}'.format(value_type=value_type))
+				raise TypeError('Bad html: expected bool, got {value_type}'.format(value_type=type(value)))
 			
 			else:
 				if value not in (0, 1):
-					raise ValueError('Bad html: expected bool, got {value_type} that is not coercible to (0, 1)'.format(value_type=value_type))
+					raise TypeError('Bad html: expected bool, got {value_type} that is not coercible to (0, 1)'.format(value_type=type(value)))
 		
-		if value and name in set(('message', 'title', 'url', 'url_title', 'device', 'callback', 'sound', 'priority', 'retry', 'expire')):
-			if name in set(('message', 'title', 'url', 'url_title', 'device', 'callback', 'sound')):
-				try:
-					value = unicode(value)
+		elif name in ('message', 'title', 'url', 'url_title', 'device', 'callback', 'sound'):
+			if value is not None:
+				if not isinstance(value, basestring):
+					raise TypeError('Bad {name}: expected string, got {type}'.format(name=name, type=type(value)))
 				
-				except ValueError:
-					raise ValueError('Bad {name}: expected string, got {type}'.format(name=name, type=type(value)))
-			
-			elif name in set(('priority', 'retry', 'expire')):
-				try:
-					value = int(value)
+				elif name == 'message' and not (0 < len(value) <= 1024):
+					raise ValueError('Bad message: must be 0-1024 characters, was {length}'.format(length=len(value)))
 				
-				except ValueError:
-					raise ValueError('Bad {name}: expected int, got {type}'.format(name=name, type=type(value)))
-			
-			if name == 'title' and len(value) > 250:
-				raise ValueError('Bad title: must be <= 250 characters, was {length}'.format(length=len(value)))
-			
-			elif name == 'message' and len(value) > 1024:
-				raise ValueError('Bad message: must be <= 1024 characters, was {length}'.format(length=len(value)))
-			
-			elif name == 'url' and len(value) > 512:
-				raise ValueError('Bad url: must be <= 512 characters, was {length}'.format(length=len(value)))
-			
-			elif name == 'url_title' and len(value) > 100:
-				raise ValueError('Bad url_title: must be <= 100 characters, was {length}'.format(length=len(value)))
-			
-			elif name == 'timestamp':
-				try:
-					if isinstance(value, datetime):
-						value = datetime_to_epoch(value)
+				if name == 'title' and len(value) > 250:
+					raise ValueError('Bad title: must be <= 250 characters, was {length}'.format(length=len(value)))
+				
+				elif name == 'url' and len(value) > 512:
+					raise ValueError('Bad url: must be <= 512 characters, was {length}'.format(length=len(value)))
+				
+				elif name == 'url_title' and len(value) > 100:
+					raise ValueError('Bad url_title: must be <= 100 characters, was {length}'.format(length=len(value)))
+				
+				elif name == 'device':
+					if not DEVICE_RE.match(value):
+						raise ValueError('Bad device: expected string matching r{pattern!r}, got {value!r}'.format(pattern=DEVICE_RE.pattern, value=value))
 					
+					elif self.user.is_authenticated:
+						if value not in self.user.devices:
+							raise ValueError('Bad device: must be in ({devices}), was {value!r}'.format(
+								devices=', '.join(repr(s) for s in sorted(self.user.devices)),
+								value=value,
+							))
 					else:
+						logger.warning('Unverified device: {ancestor} is unauthenticated, {value!r} may be bad'.format(
+							ancestor='application' if self.user.app._is_authenticated is False else 'user',
+							value=value
+						))
+				
+				elif name == 'sound':
+					if self.user.app.is_authenticated:
+						if value not in self.user.app.sounds:
+							raise ValueError('Bad sound: must be in ({sounds}), was {value!r}'.format(
+								sounds=', '.join(repr(s) for s in sorted(self.user.app.sounds.keys())),
+								value=value,
+							))
+					else:
+						logger.warning('Unverified sound: application is unauthenticated, {value!r} may be bad'.format(value=value))
+		
+		elif name == 'priority':
+			try:
+				value = int(value)
+			
+			except ValueError:
+				raise TypeError('Bad {name}: expected int, got {type}'.format(name=name, type=type(value)))
+		
+		elif name == 'timestamp':
+			if value is not None:
+				try:
+					if not isinstance(value, datetime):
 						value = epoch_to_datetime(value)
 				
 				except (TypeError, ValueError):
 					raise TypeError('Bad timestamp: expected valid int or datetime, got {value_type}.'.format(value_type=type(value)))
+		
+		elif name == 'priority':
+			try:
+				if not -2 <= int(value) <= 2:
+					raise ValueError('Bad priority: must be between -2 and 2, was {value!r}'.format(value=value))
 			
-			elif name == 'priority':
-				try:
-					if not -2 <= int(value) <= 2:
-						raise ValueError('Bad priority: must be between -2 and 2, was {value!r}'.format(value=value))
-				
-				except TypeError:
-					raise TypeError('Bad priority: expected int, got {value_type}.'.format(value_type=type(value)))
-			
-			elif name == 'sound' and value not in self.user.app.sounds:
-				raise ValueError('Bad sound: must be in ({sounds}), was {value!r}'.format(
-					sounds=', '.join(repr(s) for s in sorted(self.user.app.sounds.keys())),
-					value=value,
-				))
-			
-			elif name == 'device' and value not in self.user.devices:
-				raise ValueError('Bad device: must be in ({devices}), was {value!r}'.format(
-					devices=', '.join(repr(s) for s in sorted(self.user.devices)),
-					value=value,
-				))
+			except TypeError:
+				raise TypeError('Bad priority: expected int, got {value_type}.'.format(value_type=type(value)))
 		
 		super(Message, self).__setattr__(name, value)
 	
@@ -658,10 +727,13 @@ class Message(object):
 			'user': self.user.token,
 		}
 		
-		for kwarg in ('message', 'html', 'title', 'timestamp', 'url', 'url_title', 'device', 'priority', 'sound', 'retry', 'expire', 'callback'):
-			if hasattr(self, kwarg) and getattr(self, kwarg):
+		for kwarg in ('message', 'html', 'title', 'url', 'url_title', 'device', 'priority', 'sound', 'retry', 'expire', 'callback'):
+			if getattr(self, kwarg, None):
 				data[kwarg] = getattr(self, kwarg)
 		
+		if self.timestamp:
+			data['timestamp'] = datetime_to_epoch(self.timestamp)
+		
 		try:
 			# We've got to store this somewhere so that EmergencyMessage can check it for a receipt.
 			self._response, self.sent_at = self.user.app._request('message', data)
@@ -672,14 +744,19 @@ class Message(object):
 			
 			# This could be handled by calling {user,app}._authenticate, but that's two extra requests.
 			if 'token' in error.bad_inputs:
-				self.user.app.is_authenticated = False
-				self.user.is_authenticated = None
+				self.user.app._is_authenticated = False
+				self.user.app._sounds = None
+				self.user._is_authenticated = None
+				self.user._devices = None
 			
 			elif 'user' in error.bad_inputs:
-				self.user.is_authenticated = False
+				self.user._is_authenticated = False
+				self.user._devices = None
 		
 		else:
 			self.is_sent = True
+			self.user._is_authenticated = True
+			self.user.app._is_authenticated = True
 			self.id = self._response['request']
 		
 		return self.is_sent
@@ -737,13 +814,13 @@ class EmergencyMessage(Message):
 				value = int(value)
 			
 			except ValueError:
-				raise ValueError('Bad {name}: expected int, got {type}'.format(name=name, type=type(value)))
-		
-		if name == 'retry' and value < 30:
-			raise ValueError('Bad retry: must be >= 30, was {value}'.format(value=value))
-		
-		elif name == 'expire' and not 0 < value <= 86400:
-			raise ValueError('Bad expire: must be <= 86400 and >= 0, was {value}'.format(value=value))
+				raise TypeError('Bad {name}: expected int, got {type}'.format(name=name, type=type(value)))
+			
+			if name == 'retry' and value < 30:
+				raise ValueError('Bad retry: must be >= 30, was {value}'.format(value=value))
+			
+			elif name == 'expire' and not 0 < value <= 86400:
+				raise ValueError('Bad expire: must be 0-86400, was {value}'.format(value=value))
 		
 		super(EmergencyMessage, self).__setattr__(name, value)
 	
@@ -776,8 +853,7 @@ class EmergencyMessage(Message):
 		
 		if self.is_sent:
 			self.receipt = self._response['receipt']
-		
-		self.poll() # Poll immediately to fill attributes.
+			self.poll() # Poll immediately to fill attributes.
 		
 		return self.is_sent
 	
